﻿namespace HeuristicLab.Problems.ProgramSynthesis.Push.Expressions {
  using System;
  using Common;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Attributes;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Interpreter;
  using HeuristicLab.Problems.ProgramSynthesis.Push.Stack;

  /// <summary>
  ///     Pushes the sum of the top two items.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.+")]
  public class FloatAddExpression : PushResultExpression<double> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.FloatStack, 2,
        values => {
          var result = values[0] + values[1];

          if (double.IsPositiveInfinity(result)) return double.MaxValue;
          if (double.IsNegativeInfinity(result)) return double.MinValue;

          return result;
        });
    }
  }

  /// <summary>
  ///     Pushes the difference of the top two items; that is, the second item minus the top item.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.-")]
  public class FloatSubtractExpression : PushResultExpression<double> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.FloatStack, 2,
        values => {
          var result = values[0] - values[1];

          if (double.IsPositiveInfinity(result)) return double.MaxValue;
          if (double.IsNegativeInfinity(result)) return double.MinValue;

          return result;
        });
    }
  }

  /// <summary>
  ///     Pushes the product of the top two items.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.*")]
  public class FloatMultiplyExpression : PushResultExpression<double> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.FloatStack, 2,
        values => {
          var result = values[0] * values[1];

          if (double.IsPositiveInfinity(result)) return double.MaxValue;
          if (double.IsNegativeInfinity(result)) return double.MinValue;

          return result;
        });
    }
  }

  /// <summary>
  ///     Pushes the quotient of the top two items; that is, the second item divided by the top item.
  ///     If the top item is zero this acts as a NOOP.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT./")]
  public class FloatDivideExpression : PushResultExpression<double> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.FloatStack, 2, values => {
        var result = values[0] / values[1];

        if (double.IsPositiveInfinity(result)) return double.MaxValue;
        if (double.IsNegativeInfinity(result)) return double.MinValue;

        return result;
      }, 0);
    }
  }

  /// <summary>
  ///     Pushes the second stack item modulo the top stack item. If the top item is zero this acts as a NOOP.
  ///     The modulus is computed as the remainder of the quotient, where the quotient has first been truncated toward
  ///     negative infinity.
  ///     (This is taken from the definition for the generic MOD function in Common Lisp, which is described
  ///     for example at http://www.lispworks.com/reference/HyperSpec/Body/f_mod_r.htm.)
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.%")]
  public class FloatModuloExpression : PushResultExpression<double> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.FloatStack, 2, values => values[0] % values[1], 0);
    }
  }

  /// <summary>
  ///     Pushes the minimum of the top two items.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.MIN")]
  public class FloatMinExpression : PushResultExpression<double> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.FloatStack, 2, values => Math.Min(values[0], values[1]));
    }
  }

  /// <summary>
  ///     Pushes the maximum of the top two items.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.MAX")]
  public class FloatMaxExpression : PushResultExpression<double> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.FloatStack, 2, values => Math.Max(values[0], values[1]));
    }
  }

  /// <summary>
  ///     Pushes TRUE onto the BOOLEAN stack if the second item is less than the top item, or FALSE otherwise.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.<", StackTypes.Boolean)]
  public class FloatSmallerThanExpression : PushResultExpression<double> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.FloatStack, interpreter.BooleanStack, 2, values => values[0] < values[1]);
    }
  }

  /// <summary>
  ///     Pushes TRUE onto the BOOLEAN stack if the second item is less than the top item, or FALSE otherwise.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.<=", StackTypes.Boolean)]
  public class FloatSmallerThanOrEqualExpression : PushResultExpression<double> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.FloatStack, interpreter.BooleanStack, 2, values => values[0] < values[1] || values[0].IsAlmost(values[1]));
    }
  }

  /// <summary>
  ///     Pushes TRUE onto the BOOLEAN stack if the second item is greater than the top item, or FALSE otherwise.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.>", StackTypes.Boolean)]
  public class FloatGreaterThanExpression : PushResultExpression<double> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.FloatStack, interpreter.BooleanStack, 2, values => values[0] > values[1]);
    }
  }

  /// <summary>
  ///     Pushes TRUE onto the BOOLEAN stack if the second item is greater than or equal to the top item, or FALSE otherwise.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.>=", StackTypes.Boolean)]
  public class FloatGreaterThanOrEqualExpression : PushResultExpression<double> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.FloatStack, interpreter.BooleanStack, 2, values => values[0] > values[1] || values[0].IsAlmost(values[1]));
    }
  }

  /// <summary>
  ///     Pushes the sine of the top item.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.SIN")]
  public class FloatSineExpression : PushResultExpression<double> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.FloatStack, 1, values => Math.Sin(values[0]));
    }
  }

  /// <summary>
  ///     Pushes the cosine of the top item.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.COS")]
  public class FloatCosineExpression : PushResultExpression<double> {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      return Eval(interpreter.FloatStack, 1, values => Math.Cos(values[0]));
    }
  }

  /// <summary>
  ///     Pushes 1 if the top BOOLEAN is TRUE, or 0 if the top BOOLEAN is FALSE.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.FROMBOOLEAN", StackTypes.Boolean)]
  public class FloatFromBooleanExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.BooleanStack.Count == 0) return false;

      var condition = interpreter.BooleanStack.Pop();
      var value = condition ? 1 : 0;

      interpreter.FloatStack.Push(value);

      return true;
    }
  }

  /// <summary>
  ///     Pushes a floating point version of the top INTEGER.
  /// </summary>
  [PushExpression(StackTypes.Float, "FLOAT.FROMINTEGER", StackTypes.Integer)]
  public class FloatFromIntegerExpression : StatelessExpression {
    public override bool Eval(IInternalPushInterpreter interpreter) {
      if (interpreter.IntegerStack.Count == 0) return false;

      var value = (double)interpreter.IntegerStack.Pop();

      interpreter.FloatStack.Push(value);

      return true;
    }
  }
}